import pickle

from .base import BaseArchivesDatabase


class PicklingArchivesDatabase(BaseArchivesDatabase):
    """
    An archiving database basing off pickling things on disk

    :param path: path to the pickle. If does not exist, will be created
    """

    def on_archiving_data_sync(self, new_data) -> None:
        self.__data = new_data
        with open(self.__path, 'wb') as f_out:
            pickle.dump(self.__data, f_out, pickle.HIGHEST_PROTOCOL)

    def get_archiving_instructions(self) -> dict:
        return self.__data

    def __init__(self, path: str):
        self.__path = path
        try:
            with open(self.__path, 'rb') as f_in:
                self.__data = pickle.load(f_in)
        except (FileNotFoundError, pickle.UnpicklingError):
            self.__data = {}



from .certificate import get_root_cert, get_dev_ca_cert, get_rapid_ca_cert
from .client import SMOKDevice
from .slave import SlaveDevice

__all__ = ['SMOKDevice', 'SlaveDevice', 'get_root_cert', 'get_dev_ca_cert',
           'get_rapid_ca_cert']



import enum
import typing as tp

from satella.coding.transforms import clip

from smok.exceptions import OperationFailedError

PathpointValueType = tp.Union[int, float, str]

ValueOrExcept = tp.Union[PathpointValueType, OperationFailedError]


class PathpointType(enum.Enum):
    B = 'B'  #: binary, 0 or 1
    W = 'W'  #: unsigned 16-bit
    w = 'w'  #: signed 16-bit
    f = 'f'  #: IEEE 754 float
    d = 'd'  #: IEEE 754 double
    u = 'u'  #: an Unicode string

    @staticmethod
    def get_type(path: str) -> 'PathpointType':
        """
        Return the type of given pathpoint

        :return: type of the pathpoint
        """
        if path[0] == 'r':
            return PathpointType(path[1])
        else:
            return PathpointType(path[0])


def to_type(value: PathpointValueType, type_: PathpointType) -> PathpointValueType:
    """
    Coerces a value to target pathpoint type

    :param value: value to coerce
    :param type_: type of target pathpoint
    :return: a coerced value
    """
    if type_ == PathpointType.B:
        v = bool(int(value))
    elif type_ == PathpointType.W:
        v = clip(int(value), 0, 65535)
    elif type_ == PathpointType.w:
        v = clip(int(value), -32768, 32767)
    elif type_ in (PathpointType.f, PathpointType.d):
        v = float(value)
    elif type_ == PathpointType.u:
        v = str(value)
    else:
        raise ValueError('Invalid type given')
    return v



import logging
import base64
import gzip
from logging import LogRecord, Handler

import ujson
from satella.coding import Monitor
from satella.coding.concurrent import SequentialIssuer
from satella.instrumentation import Traceback, frame_from_traceback
from satella.instrumentation.memory import MemoryPressureManager
from satella.time import time_us
from smok.threads.log_publisher import MAX_LOG_BUFFER_SIZE

__all__ = ['SMOKLogHandler']

logger = logging.getLogger(__name__)


class SMOKLogHandler(Handler, Monitor):
    """
    A logging handler for SMOK

    :param device: SMOKDevice that this logger is attached to
    :param service_name: name of the service
    """

    def __init__(self, device: 'SMOKDevice', service_name: str):
        super().__init__()
        Monitor.__init__(self)

        self.service_name = service_name
        self.device = device
        # So timestamps inserted will be monotonically increasing
        self.timestamper = SequentialIssuer(time_us())

        mpm = MemoryPressureManager()
        self.mem_callback = mpm.register_on_entered_severity(2)(self.prune_the_queue)

    def __del__(self):
        self.mem_callback.cancel()

    def prune_the_queue(self):
        while self.device.log_publisher.queue.qsize():
            self.device.log_publisher.queue.get()
        logger.error('Pruned the log queue thanks to low memory condition')

    def record_to_json(self, record: LogRecord):
        ts = self.timestamper.no_less_than(time_us())

        try:
            msg = self.format(record)
        except (TypeError, ValueError):
            try:
                msg = record.message + ' ' + ','.join(map(repr, record.args))
            except AttributeError: # record nas no attribute .message
                try:
                    msg = str(record)
                except TypeError:
                    msg = '<empty>'

        if len(msg) > 1000:
            msg_content = base64.b64encode(
                gzip.compress(msg.encode('utf8'), compresslevel=8)).decode('utf8')
            msg = {'encoding': 'base64-gzip', 'content': msg_content}

        dct = {
            'service': self.service_name,
            'when': ts,
            'message': msg,
            'level': record.levelno,
        }

        if record.exc_info:
            f = frame_from_traceback(record.exc_info[2])
            tb = Traceback(f)
            try:
                tb_json = base64.b64encode(
                    gzip.compress(ujson.dumps(tb.to_json()).encode('utf-8'), 9)).decode('utf-8')
            except OverflowError:
                return None
            dct.update(exception_text=tb.pretty_format(),
                       exception_traceback=tb_json)

        return dct

    def emit(self, record: LogRecord) -> None:
        if self.device.log_publisher.queue.qsize() < MAX_LOG_BUFFER_SIZE:
            le = self.record_to_json(record)
            if le is not None:
                self.device.log_publisher.queue.put(le)

    def close(self) -> None:
        pass



import typing as tp
import os
import pickle

from .base import BasePredicateDatabase


class PicklingPredicateDatabase(BasePredicateDatabase):
    def __init__(self, path: str):
        self.__path = path
        self.__predicates = {}
        if os.path.exists(self.__path):
            with open(self.__path, 'rb') as f_in:
                self.__predicates = pickle.load(f_in)

    def __sync(self):
        with open(self.__path, 'wb') as f_out:
            pickle.dump(self.__predicates, f_out, pickle.HIGHEST_PROTOCOL)

    def get_all_predicates(self) -> tp.List[tp.Dict]:
        return list(self.__predicates.values())

    def update_predicate(self, v: tp.Dict) -> None:
        self.__predicates[v['predicate_id']] = v
        self.__sync()

    def set_new_predicates(self, v: tp.List[tp.Dict]):
        self.__predicates = {}
        for dct in v:
            self.__predicates[dct['predicate_id']] = dct
        self.__sync()



import typing as tp
from abc import ABCMeta, abstractmethod


class BaseMetadataDatabase(metaclass=ABCMeta):
    """
    Base class for metadata databases
    """

    @abstractmethod
    def put_plain(self, key: str, value: str, timestamp: tp.Optional[float] = None) -> None:
        """
        A plain value was just written.

        :param key: key of the write
        :param value: value of the write
        :param timestamp: timestamp of the write, default is current
        """

    @abstractmethod
    def get_plain(self, key: str) -> str:
        """
        Return a value for given plain key

        :param key: key to give a value for
        :raises KeyError: key not found, ask the server
        """

    @abstractmethod
    def update_plain(self, key: str, value: str, timestamp: float) -> None:
        """
        On an update plain request from the server

        :param key: key of the key
        :param value: value of the key
        :param timestamp: timestamp at which time it was last written
        """

    @abstractmethod
    def delete_plain(self, key: str) -> None:
        """
        Delete a particular key for plain database

        No-op on key not being in the database.

        :param key: key to delete
        """

    @abstractmethod
    def get_all_plain(self) -> tp.Iterator[tp.Tuple[str, str, float]]:
        """
        Get all plain keys in the database

        :return: an iterator of tuple(key, value, timestamp of the write)
        """



import logging
import typing as tp

from .base import BaseStatistic

logger = logging.getLogger(__name__)


class UndefinedStatistic(BaseStatistic):
    """A placeholder statistic for undefined statistics"""
    statistic_name = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.args = args
        self.kwargs = kwargs

    def initialize(self, cls: tp.Type[BaseStatistic]):
        return cls(*self.args, **self.kwargs)

    def check_device(self):
        for predicate_id in self.device.predicates:
            pred = self.device.predicates[predicate_id]
            if isinstance(pred, UndefinedStatistic):
                statistic_name = pred.statistic
                base_class = self.device.statistic_registration.try_match(pred.statistic,
                                                                          pred.configuration)
                if base_class is not None:
                    logger.info(f'Initialized missing predicate ID %s statistic %s', predicate_id,
                                statistic_name)
                    self.device.predicates[predicate_id] = pred.initialize(base_class)

    def on_tick(self) -> None:
        # check whether the correct statistic classes were loaded and substitute
        # undefined statistic with these
        self.check_device()



from .base import BasePathpointDatabase, BaseDataToSynchronize
from .pickling import PicklingPathpointDatabase

__all__ = ['BasePathpointDatabase', 'BaseDataToSynchronize',
           'PicklingPathpointDatabase']



import logging
import time
import typing as tp
import weakref
from abc import ABCMeta, abstractmethod
from datetime import datetime
from satella.coding.structures import OmniHashableMixin

from smok.predicate.event import Color, Event

logger = logging.getLogger(__name__)


class Time(OmniHashableMixin):
    """
    A representation of a time during a weekday

    :ivar day_of_week: day of week, as per ISO 8601
    :ivar hour: a hour, according to a 24-hour clock
    :ivar minute: a minute
    """
    _HASH_FIELDS_TO_USE = ('day_of_week', 'hour', 'minute')
    __slots__ = ('day_of_week', 'hour', 'minute')

    def __init__(self, day_of_week: int, hour: int, minute: int):
        self.day_of_week = day_of_week
        self.hour = hour
        self.minute = minute

    @classmethod
    def from_json(cls, dct: dict) -> 'Time':
        return Time(dct['day'], dct['hour'], dct['minute'])

    def to_tuple(self) -> tp.Tuple[int, int, int]:
        return self.day_of_week, self.hour, self.minute


class DisabledTime(OmniHashableMixin):
    """
    Class marking a period during a week

    :ivar start: when this period starts
    :ivar stop: when this period stops
    """
    _HASH_FIELDS_TO_USE = ('start', 'stop')
    __slots__ = ('start', 'stop')

    def __init__(self, start: Time, stop: Time):
        self.start = start
        self.stop = stop

    @classmethod
    def from_json(cls, x: dict) -> 'DisabledTime':
        return DisabledTime(Time.from_json(x['start']), Time.from_json(x['stop']))

    def is_in_time(self, t: datetime) -> bool:
        """
        Check whether provided time is inside the range of this silencing period.

        :return: True if the time is inside
        """
        return self.start.to_tuple() <= (t.isoweekday(), t.hour, t.minute) \
               <= self.stop.to_tuple()


class BaseStatistic(metaclass=ABCMeta):
    """
    Base class for your own predicates.

    :ivar device: a weak reference to the SMOKDevice
    :ivar predicate_id: ID of the predicate
    :ivar verbose_name: Human-readable name of this predicate
    :ivar silencing: periods during which the predicate shouldn't generate alerts
    :ivar configuration: a dictionary containing the predicate's configuration
    :ivar group: notification group
    :ivar state: state of the predicate, persisted between calls (picklable)
    :cvar statistic_name: name of this statistic
    """

    def __init__(self, device: 'SMOKDevice', predicate_id: str, verbose_name: str,
                 silencing: tp.List[DisabledTime],
                 configuration: tp.Optional[dict], statistic: tp.Optional[str] = None,
                 group: str = 'B', state=None, **kwargs):
        self.device = weakref.proxy(device)
        self.predicate_id = predicate_id
        self.verbose_name = verbose_name
        self.silencing = silencing
        self.configuration = configuration
        self.statistic = statistic
        self.group = group
        self.state = state
        self.kwargs = kwargs

    def to_kwargs(self) -> dict:
        r = {'group': self.group, 'predicate_id': self.predicate_id,
             'verbose_name': self.verbose_name, 'silencing': self.silencing,
             'configuration': self.configuration, 'statistic': self.statistic,
             'state': self.state, **self.kwargs}
        if 'device' in r:
            del r['device']
        return r

    @abstractmethod
    def on_tick(self) -> None:
        """
        Called about each 20 seconds by the communicator thread. This should commence any required
        analyses.

        :attr:`state` is loaded before this call and persisted after it finishes
        """

    def _call_method(self, method: str, *args, **kwargs):
        """
        Securely call a method of this object with provided arguments
        """
        try:
            self.state = self.device.evt_database.get_cache(self.predicate_id)
        except KeyError:
            self.state = None
        try:
            getattr(self, method)(*args, **kwargs)
        finally:
            self.device.evt_database.set_cache(self.predicate_id, self.state)

    def close_event(self, event: Event) -> None:
        """
        Close an event

        :param event: event to close
        """
        if not event.is_closed():
            event.ended_on = time.time()
        self.device.evt_database.close_event(event)

    def open_event(self, msg: str, color: Color) -> tp.Optional[Event]:
        """
        Open an event.

        This automatically checks for current silencing effect, and will return None
        if current time indicates that the event should be silenced.

        :param msg: extra message for the event
        :param color: color of the event
        :return: an Event if silencing is not in effect, else None
        """
        # Check the silencing rules
        localtime = self.device.get_local_time()
        for silencing_period in self.silencing:
            if silencing_period.is_in_time(localtime):
                return

        message = self.verbose_name
        if msg:
            message = '%s: %s' % (message, msg)

        evt = Event(None, None, None, color, False, self.statistic, self.group,
                    message, None, {'predicate_id': self.predicate_id})
        self.device.evt_database.add_event(evt)
        return evt

    def on_group_changed(self, new_group: str) -> None:
        """
        Called upon group changing. This should assign the changed verbose name.

        Called by communicator thread.

        :param new_group: new verbose name
        """
        self.group = new_group

    def on_verbose_name_changed(self, new_verbose_name: str) -> None:
        """
        Called upon verbose name changing. This should assign the changed verbose name.

        Called by communicator thread.

        :param new_verbose_name: new verbose name
        """
        self.verbose_name = new_verbose_name

    def on_configuration_changed(self, new_config: dict) -> None:
        """
        Called upon configuration changing. This should assign the changed configuration.

        Called by communicator thread.

        :param new_config: new configuration
        """
        self.configuration = new_config

    def on_silencing_changed(self, new_silencing: tp.List[DisabledTime]) -> None:
        """
        Called upon silencing rules changing. This should assign the changed silencing.

        Called by communicator thread.

        :param new_silencing: new silencing
        """
        self.silencing = new_silencing

    def on_offline(self) -> None:
        """
        Called when the predicate is disabled or deleted.
        After that, this class will be destroyed, and if the predicate gets enabled again,
        a new instance will be created.

        Called by communicator thread.
        """



import ujson as ujson
from satella.coding.decorators import retry

from smok.basics import SlaveDeviceInfo
from smok.exceptions import ResponseError


class SlaveDevice:
    """
    A device that's attached to it's master

    :ivar device_id: slave device ID (str)
    :ivar configuration: a string configuring this device (str)
    :ivar responsible_service: should always be "rapid" (str)
    :ivar master_controller: device ID of it's master device (str)
    """
    __slots__ = ('device_id', 'sd', '__linkstate', '__instrumentation',
                 'configuration', 'responsible_service', 'master_controller')

    def __init__(self, sd: 'SMOKDevice', data: SlaveDeviceInfo):
        self.device_id = data.device_id
        self.configuration = data.configuration
        self.responsible_service = data.responsible_service
        self.master_controller = data.master_controller
        self.sd = sd
        self.__linkstate = None
        self.__instrumentation = None

    @property
    def linkstate(self) -> dict:
        """
        An (initially empty) dictionary of format (OpenAPI 3.0):

        .. code-block:: yaml

            type: object
            properties:
                status:
                    type: boolean
                    description: Is the link OK?
                failed_devices:
                    type: array
                    items:
                        type: integer
                        description: Address of the failed device
            required:
                - status

        That tells that status of the device's link, for usage by predicates
        detecting it's failures.

        :returns: current link state
        """
        if self.__linkstate is None:
            resp = self.sd.api.get('/v1/device/instrumentation/%s' % (self.device_id,))
            try:
                linkstate = ujson.loads(resp['linkstate'])
            except ValueError:
                linkstate = {}
            self.__linkstate = linkstate
            self.__instrumentation = resp['instrumentation']
        return self.__linkstate

    @linkstate.setter
    @retry(3, exc_classes=ResponseError)
    def linkstate(self, v: dict) -> None:
        assert 'status' in v, 'Status not in dictionary!'
        self.sd.api.patch('/v1/device/instrumentation/%s' % (self.device_id,), json={
            'linkstate': ujson.dumps(v)
        })
        self.__linkstate = v

    @property
    def instrumentation(self) -> str:
        """
        An arbitrary sequence of characters telling this slave's condition.

        :return: current instrumentation
        """
        if self.__instrumentation is None:
            resp = self.sd.api.get('/v1/device/instrumentation/%s' % (self.device_id,))
            self.__linkstate = resp['linkstate']
            self.__instrumentation = resp['instrumentation']
        return self.__instrumentation

    @instrumentation.setter
    @retry(3, exc_classes=ResponseError)
    def instrumentation(self, v: str) -> None:
        self.sd.api.patch('/v1/device/instrumentation/%s' % (self.device_id,), json={
            'instrumentation': v
        })
        self.__instrumentation = v



from smok.extras import BaseArchivesDatabase


class InMemoryArchivesDatabase(BaseArchivesDatabase):
    def on_archiving_data_sync(self, new_data: dict) -> None:
        self.data = new_data

    def get_archiving_instructions(self) -> dict:
        return self.data

    def __init__(self):
        self.data = {}



import typing as tp

from satella.coding.structures import ReprableMixin, HashableIntEnum
from satella.json import JSONAble


class Environment(HashableIntEnum):
    """
    An environment in which this device runs
    """
    PRODUCTION = 0  #: production
    STAGING = 1  #: testing environment
    LOCAL_DEVELOPMENT = 2  #: CI or local development


class StorageLevel(HashableIntEnum):
    """
    A storage level defines how long is the pathpoint kept at SMOK server.
    """
    PERMANENT = 0  #: hold all values indefinitely
    TREND = 1  #: values at most 2 weeks old will be kept

    def __or__(self, other: 'StorageLevel') -> 'StorageLevel':
        """
        Get the more permissive
        """
        return StorageLevel(self.value | other.value)


class SlaveDeviceInfo(JSONAble, ReprableMixin):
    """
    Information about a slave device attached to primary device

    :ivar device_id: slave device ID (str)
    :ivar master_controller: ID of master device (str)
    :ivar responsible_service: service responsible for this device, mostly "rapid" (str)
    :ivar configuration: a string containing configuration for this device (str)
    """
    _REPR_FIELDS = ('device_id', 'master_controller', 'responsible_service',
                    'configuration')
    __slots__ = ('device_id', 'master_controller', 'responsible_service',
                 'configuration')

    def __init__(self, device_id: str, master_controller: str, responsible_service: str,
                 configuration: str):
        self.device_id = device_id
        self.master_controller = master_controller
        self.responsible_service = responsible_service
        self.configuration = configuration

    @classmethod
    def from_json(cls, dct: dict) -> 'SlaveDeviceInfo':
        return SlaveDeviceInfo(dct['device_id'], dct['master_controller'],
                               dct['responsible_service'], dct['configuration'])

    def to_json(self) -> dict:
        return {
            'device_id': self.device_id,
            'configuration': self.configuration,
            'master_controller': self.master_controller,
            'responsible_service': self.responsible_service
        }


class DeviceInfo(ReprableMixin):
    """
    A class holding device information

    :ivar device_id: device ID of the device (str)
    :ivar slaves: list of :class:`smokclient.basics.SlaveDeviceInfo` containing info
        about the slave devices (tp.List[:class:`~smok.basics.SlaveDeviceInfo`])
    :ivar facets: a set of strings, contains interfaces that access to this device is allowed for
        (tp.Set[str])
    :ivar timezone: local timezone of this device, in accordance with tzdata (str)
    :ivar language: language of this device, according to ISO639-1 (str)
    :ivar units: either `metric` or `imperial`, units used on this device (str)
    :ivar verbose_name: human-readable name of this device (str)
    """
    _REPR_FIELDS = ('device_id', 'facets', 'language', 'timezone', 'units',
                    'slaves')
    __slots__ = ('device_id', 'slaves', 'facets', 'language', 'timezone', 'units',
                 'verbose_name')

    def __init__(self, device_id: str, facets: tp.Set[str], language: str, timezone: str,
                 units: str, verbose_name: str, slaves: tp.List[SlaveDeviceInfo]):
        self.device_id = device_id
        self.facets = facets
        self.language = language
        self.timezone = timezone
        self.units = units
        self.verbose_name = verbose_name
        self.slaves = slaves

    @classmethod
    def from_json(cls, dct: dict) -> 'DeviceInfo':
        ccon = dct['culture_context']
        return DeviceInfo(dct['device_id'], set(dct['facets']), ccon['language'],
                          ccon['timezone'], ccon['units'], dct['verbose_name'],
                          [SlaveDeviceInfo.from_json(y) for y in dct['slave_devices']])



import typing as tp
from abc import ABCMeta, abstractmethod

from smok.macro import Macro


class BaseMacroDatabase(metaclass=ABCMeta):

    @abstractmethod
    def get_done_macros(self) -> tp.Iterator[tp.Tuple[str, int]]:
        """
        Get a list of macros that were done, but the server wasn't told about that yet.

        Called by communicator thread

        :return: an iterator of a tuple (macro_id, executed timestamps)
        """

    @abstractmethod
    def notify_macro_synced(self, macro_id: str, timestamp: int) -> None:
        """
        Notify the DB that server has received notification about executing given macro.

        Called by communicator thread
        """

    @abstractmethod
    def notify_macro_executed(self, macro_id: str, timestamp: int) -> None:
        """
        Notify the DB that the device has executed a given macro.

        This should update the database's macro definition as well.

        Called by archive & macro thread
        """

    @abstractmethod
    def set_macros(self, macros: tp.List[Macro]) -> None:
        """
        Set current list of macros to execute (have been not yet executed)

        :param macros: list of macro to execute
        """

    @abstractmethod
    def get_macros(self) -> tp.List[Macro]:
        """
        Get a list of macros to execute
        """

    @abstractmethod
    def checkpoint(self) -> None:
        """
        Called about each 20 seconds by communicator thread
        """



import typing as tp
from abc import abstractmethod, ABCMeta


class BaseArchivesDatabase(metaclass=ABCMeta):
    """
    A base class for databases with archiving instructions.
    """

    @abstractmethod
    def get_archiving_instructions(self) -> tp.Dict[int, tp.List[str]]:
        """
        Return currently stored archiving instructions, or an empty dict if none available.

        Called by `ArchivingAndMacroThread`

        :return: a dictionary of (archiving_interval::int => list of pathpoints to archive)
        """

    @abstractmethod
    def on_archiving_data_sync(self, new_data: tp.Dict[int, tp.List[str]]) -> None:
        """
        Called on new archiving data instructions.

        Called by `ArchivingAndMacroThread`

        :param new_data: a dictionary of (archiving_interval::int => list of pathpoints to archive)
        """



import logging
import typing as tp

from ngtt.protocol import NGTTHeaderType

logger = logging.getLogger(__name__)


class Order:
    """
    An order received from the server.

    Can be confirmed via :meth:`~ngtt.orders.Order.acknowledge` to signal to the server
    that it's been processed.
    """
    __slots__ = ('data', 'tid', 'sock', 'confirmed')

    def __init__(self, data: tp.List[dict], tid: int, sock: 'NGTTSocket'):
        self.data = data
        self.tid = tid
        self.sock = sock
        self.confirmed = False

    def __repr__(self) -> str:
        return '<Order(%s, %s)>' % (repr(self.data), self.tid)

    def __str__(self) -> str:
        return repr(self)

    def acknowledge(self):
        """
        Signal to the server that the order has been processed
        """
        if not self.confirmed:
            self.confirm_with(NGTTHeaderType.ORDER_CONFIRM)

    def confirm_with(self, c_type: NGTTHeaderType):
        self.sock.send_frame(self.tid, c_type)
        self.confirmed = True

    def nack(self):
        """
        Signal the server that the order has been rejected
        """
        if not self.confirmed:
            self.confirm_with(NGTTHeaderType.ORDER_REJECT)



import typing as tp
from abc import ABCMeta, abstractmethod

from smok.predicate.event import Event


class BaseEventSynchronization(metaclass=ABCMeta):
    __slots__ = ()

    @abstractmethod
    def get_events(self) -> tp.List[Event]:
        """
        :return: a list of events to synchronize
        """

    @abstractmethod
    def acknowledge(self, *uuids: str) -> None:
        """
        Called by the communicator, when sync succeeds

        :param uuids: UUIDs assigned to events"""

    @abstractmethod
    def negative_acknowledge(self) -> None:
        """Called by the communicator, when sync fails"""


class BaseEventDatabase(metaclass=ABCMeta):

    @abstractmethod
    def checkpoint(self) -> None:
        """
        Called by the communicator thread, once every about 20 seconds.
        May be called much more often, it's the function responsibility to throttle.
        """

    @abstractmethod
    def get_open_events(self) -> tp.Iterator[Event]:
        """
        :return: an iterator with all open events
        """

    @abstractmethod
    def get_all_events(self) -> tp.Iterator[Event]:
        """
        :return: all events kept in the database
        """

    @abstractmethod
    def close_event(self, event: Event) -> None:
        """
        Close provided event

        :param event: event to close
        """

    @abstractmethod
    def add_event(self, event: Event) -> None:
        """
        Register a new event in the database.

        Can be called by any thread.

        :param event: event to register
        """

    @abstractmethod
    def get_events_to_sync(self) -> tp.Optional[BaseEventSynchronization]:
        """
        At most a single instance of BaseEventSynchronization will be alive at a time.

        :return: object to sync, or None if there's nothing to sync.
        """

    @abstractmethod
    def set_cache(self, predicate_id: str, cache) -> None:
        """
        Store predicate's internal data. Do it in a way that will survive restarts.
        """

    @abstractmethod
    def get_cache(self, predicate_id: str) -> tp.Any:
        """
        Return predicate's internal data

        :raises KeyError: predicate internal data not found
        """

    @abstractmethod
    def on_predicate_deleted(self, predicate_id: str) -> None:
        """
        Called when a predicate is deleted.

        Called by communicator thread.

        :param predicate_id: ID of the predicate that was deleted
        """



import logging
import queue
import time
import typing as tp

import ujson
from satella.coding import silence_excs, for_argument, log_exceptions, wraps
from satella.coding.concurrent import TerminableThread, Condition
from satella.coding.decorators import retry
from satella.coding.transforms import jsonify
from satella.exceptions import WouldWaitMore
from satella.time import measure

from smok.basics import StorageLevel
from smok.exceptions import ResponseError

from smok.extras import BasePathpointDatabase
from smok.pathpoint.orders import sections_from_list
from smok.pathpoint.pathpoint import Pathpoint
from smok.predicate import DisabledTime
from smok.predicate.undefined import UndefinedStatistic
from smok.sensor import Sensor
from smok.sync_workers.base import SyncError

logger = logging.getLogger(__name__)

SENSORS_SYNC_INTERVAL = 300


def profile(fun):
    @wraps(fun)
    def inner(*args, **kwargs):
        with measure() as measurement:
            y = fun(*args, **kwargs)
        logger.debug('Running %s took %s seconds', fun, measurement())
        return y
    return inner


@for_argument(returns=jsonify)
def pathpoints_to_json(pps: tp.Iterable[Pathpoint]) -> list:
    output = []
    for pp in pps:
        output.append({'path': pp.name,
                       'storage_level': pp.storage_level})
    return output


COMMUNICATOR_INTERVAL = 20
PREDICATE_SYNC_INTERVAL = 300
BAOB_SYNC_INTERVAL = 60 * 60  # an hour


def redo_data(data):
    """
    Alter the data received from the backend to our way
    """
    output = []
    for pp in data:
        values = []
        for ts in pp['values']:
            if isinstance(ts, dict):
                if 'error_code' in ts:
                    values.append([False, ts['timestamp'], ts['error_code']])
                else:
                    values.append([ts['timestamp'], ts['value']])
            else:
                values.append(ts)
        output.append({'path': pp['path'],
                       'values': values})
    return output


class CommunicatorThread(TerminableThread):
    def __init__(self, device: 'SMOKClient', order_queue: queue.Queue,
                 data_to_sync: BasePathpointDatabase, dont_obtain_orders: bool,
                 dont_do_baobs: bool, dont_do_pathpoints: bool,
                 dont_do_predicates: bool,
                 dont_sync_sensor_writes: bool,
                 startup_delay: float):
        super().__init__(name='order getter')
        self.dont_sync_sensor_writes = dont_sync_sensor_writes
        self.device = device
        self.startup_delay = startup_delay
        self.dont_do_pathpoints = dont_do_pathpoints
        self.dont_do_baobs = dont_do_baobs
        self.dont_obtain_orders = dont_obtain_orders
        self.queue = order_queue
        self.dont_do_predicates = dont_do_predicates
        self.data_to_sync = data_to_sync
        self.last_sensors_synced = 0
        self.last_predicates_synced = 0
        self.data_to_update = Condition()
        self.last_baob_synced = 0

    @profile
    def tick_predicates(self) -> None:
        for predicate in self.device.predicates.values():
            kwargs = predicate.to_kwargs()
            # noinspection PyProtectedMember
            predicate._call_method('on_tick')
            new_kwargs = predicate.to_kwargs()
            if kwargs != new_kwargs:
                self.device.pred_database.update_predicate(new_kwargs)

    @retry(3, ResponseError)
    @profile
    def sync_events(self) -> None:
        evt_to_sync = self.device.evt_database.get_events_to_sync()
        if evt_to_sync is None:
            return
        try:
            json = [evt.to_json() for evt in evt_to_sync.get_events()]
            resp = self.device.api.post('/v1/device/alarms',
                                        json=json)
            evt_to_sync.acknowledge(*(data['uuid'] for data in resp))
            self.device.on_successful_sync()
        except ResponseError as e:
            if e.is_no_link():
                self.device.on_failed_sync()
            logger.error('Failure syncing events: %s', e, exc_info=e)
            evt_to_sync.negative_acknowledge()
            raise

    @retry(3, ResponseError)
    @profile
    def sync_predicates(self) -> None:
        try:
            resp = self.device.api.get('/v1/device/predicates')
        except ResponseError as e:
            if e.is_no_link():
                self.device.on_failed_sync()
            raise

        predicates_found = set()
        for predicate_dict in resp:
            predicate_id = predicate_dict['predicate_id']
            predicates_found.add(predicate_id)
            if not predicate_dict['online']:
                if predicate_id in self.device.predicates:
                    self.device.predicates[predicate_id].on_offline()
                    del self.device.predicates[predicate_id]
            else:
                silencing = [DisabledTime.from_json(dct) for dct in
                             predicate_dict.get('silencing', ())]

                if predicate_id not in self.device.predicates:
                    stat_name = predicate_dict['statistic']
                    cfg = predicate_dict['configuration']
                    base_class = self.device.statistic_registration.try_match(stat_name, cfg)
                    if base_class is None:
                        base_class = UndefinedStatistic
                    predicate = base_class(self.device, predicate_id,
                                           predicate_dict['verbose_name'],
                                           silencing, cfg, stat_name)
                    self.device.predicates[predicate_id] = predicate
                else:
                    stat = self.device.predicates[predicate_id]
                    config = predicate_dict['configuration']
                    verbose_name = predicate_dict['verbose_name']
                    group = predicate_dict['group']
                    if stat.configuration != config:
                        stat.on_configuration_changed(config)
                    if stat.silencing != silencing:
                        stat.on_silencing_changed(silencing)
                    if stat.verbose_name != verbose_name:
                        stat.on_verbose_name_changed(verbose_name)
                    if stat.group != group:
                        stat.on_group_changed(group)

                predicate_dict['silencing'] = silencing

        predicates_to_delete = set(self.device.predicates.keys()) - predicates_found
        for predicate_id in predicates_to_delete:
            self.device.predicates[predicate_id].on_offline()
            del self.device.predicates[predicate_id]

        if not self.last_predicates_synced:
            self.device.ready_lock.release()
        self.last_predicates_synced = time.monotonic()
        self.db_sync_predicates()
        self.device.on_successful_sync()

    @profile
    def db_sync_predicates(self):
        lst = [predicate.to_kwargs() for predicate in self.device.predicates.values()]
        self.device.pred_database.set_new_predicates(lst)

    @retry(3, ResponseError)
    @profile
    def sync_sensors(self) -> None:
        try:
            resp = self.device.api.get('/v1/device/sensors')

            self.device.sensor_database.on_sensors_sync(
                [Sensor.from_json(self.device, data) for data in resp])
            self.last_sensors_synced = time.monotonic()
            self.device.on_successful_sync()
        except ResponseError as e:
            if e.is_no_link():
                self.device.on_failed_sync()
            raise

    @retry(3, ResponseError)
    @profile
    def fetch_orders(self) -> None:
        try:
            resp = self.device.api.post('/v1/device/orders')
            if resp:
                for section in sections_from_list(resp):
                    self.queue.put(section)
            self.device.on_successful_sync()
        except ResponseError as e:
            if e.is_no_link():
                self.device.on_failed_sync()
            raise

    @retry(3, ResponseError)
    @profile
    def sync_sensor_writes(self) -> None:
        sync = self.device.sensor_write_database.get_sw_sync()
        if not sync:
            return
        try:
            self.device.api.put('/v1/device/sensor/write_log', json=sync.to_json())
            sync.ack()
            self.device.on_successful_sync()
        except ResponseError as e:
            if e.is_no_link():
                self.device.on_failed_sync()
            if not e.is_clients_fault():
                logger.debug('Failed to sync sensor writes: %s', e, exc_info=e)
                sync.nack()
                raise
            else:
                logger.warning('Got HTTP %s on sync sensor writes, acking', e.status_code)
                sync.ack()

    @retry(3, SyncError)
    @profile
    def sync_data(self) -> None:
        sync = self.data_to_sync.get_data_to_sync()
        if sync is None:
            logger.debug('Got nothing to sync (1)')
            return
        try:
            data = sync.to_json()
            if not data:
                sync.acknowledge()
                logger.debug('Got nothing to sync (2)')
                return
            self.device.sync_worker.sync_pathpoints(redo_data(data))
            sync.acknowledge()
            self.device.on_successful_sync()
            logger.debug('Successfully synced %s elements', len(data))
        except SyncError as e:
            if e.is_no_link():
                self.device.on_failed_sync()
            if not e.is_clients_fault():
                sync.negative_acknowledge()
            else:
                logger.warning('Got HTTP %s while syncing pathpoint data. '
                               'Assuming is it damaged and marking as synced', e.status_code)
                sync.acknowledge()

    @profile
    def sync_baob(self) -> None:
        self._sync_baob()
        self.device.baobs_loaded = True

    @retry(3, ResponseError)
    def _sync_baob(self) -> None:
        try:
            keys = self.device.baob_database.get_all_keys()
            data = []
            for key in keys:
                try:
                    data.append({'key': key,
                                 'version': self.device.baob_database.get_baob_version(key)})
                except KeyError:
                    logger.error('Got key %s but the DB tells us that it does not exist', key)
                    continue
            data = self.device.api.post('/v1/device/baobs', json=data)

            for key_to_delete in data['should_delete']:
                self.device.baob_database.delete_baob(key_to_delete)

            for key_to_download in data['should_download']:
                resp, headers = self.device.api.get(f'/v1/device/baobs/{key_to_download}',
                                                    direct_response=True)
                self.device.baob_database.set_baob_value(key_to_download, resp,
                                                         int(headers['X-SMOK-BAOB-Version']))
                if self.last_baob_synced:
                    self.device.on_baob_updated(key_to_download)

            for key_to_upload in data['should_upload']:
                self.device.api.put(f'/v1/device/baobs/{key_to_upload}', files={
                    'file': self.device.baob_database.get_baob_value(key_to_upload),
                    'data': ujson.dumps(
                        {'version': self.device.baob_database.get_baob_version(
                            key_to_upload)}).encode(
                        'utf8')
                })
            self.last_baob_synced = time.monotonic()
            self.device.on_successful_sync()
        except ResponseError as e:
            if e.is_no_link():
                self.device.on_failed_sync()
            raise

    @retry(3, ResponseError)
    @profile
    def sync_pathpoints(self) -> None:
        pps = self.device.pathpoints.copy_and_clear_dirty()
        data = pathpoints_to_json(pps.values())
        try:
            resp = self.device.api.put('/v1/device/pathpoints', json=data)
            for pp in resp:
                name = pp['path']
                if name.startswith('r'):  # Don't use reparse pathpoints
                    continue
                with silence_excs(KeyError):
                    pathpoint = self.device.provide_unknown_pathpoint(name, StorageLevel(
                        pp.get(
                            'storage_level',
                            1)))
                    stor_level = StorageLevel(pp.get('storage_level', 1))
                    if stor_level != pathpoint.storage_level:
                        pathpoint.on_new_storage_level(stor_level)
            self.device.on_successful_sync()
        except ResponseError as e:
            if e.is_no_link():
                self.device.on_failed_sync()
            self.device.pathpoints.update(pps)
            self.device.pathpoints.dirty = True
            raise

    def prepare(self) -> None:
        # Give the app a moment to prepare and define it's pathpoints
        self.safe_sleep(self.startup_delay)

    @profile
    def wait(self, time_taken: float):
        time_to_wait = COMMUNICATOR_INTERVAL - time_taken
        while time_to_wait > 0.1 and not self.terminating:  # for float roundings
            try:
                ttw = min(time_to_wait, 5)
                self.data_to_update.wait(timeout=ttw)
                return
            except WouldWaitMore:
                time_to_wait -= ttw

    @log_exceptions(logger, logging.ERROR)
    def loop(self) -> None:
        with measure() as measurement:
            # Synchronize the data
            monotime = time.monotonic()

            if self.device.allow_sync:
                if not self.dont_do_pathpoints:
                    self.sync_data()

                    # Synchronize the pathpoints
                    if self.device.pathpoints.dirty:
                        self.sync_pathpoints()

                    # Synchronize sensors
                    if monotime - self.last_sensors_synced > SENSORS_SYNC_INTERVAL:
                        self.sync_sensors()

                # Synchronize predicates
                if not self.dont_do_predicates:
                    if monotime - self.last_predicates_synced > PREDICATE_SYNC_INTERVAL:
                        self.sync_predicates()

                # Fetch the BAOBs
                if not self.dont_do_baobs:
                    if monotime - self.last_baob_synced > BAOB_SYNC_INTERVAL:
                        self.sync_baob()
                if not self.dont_sync_sensor_writes:
                    self.sync_sensor_writes()

                # Fetch the orders
                if not self.dont_obtain_orders and not self.device.sync_worker.has_async_orders:
                    self.fetch_orders()

                if not self.dont_do_predicates:
                    # Tick the predicates
                    self.tick_predicates()

                    # Sync the events
                    self.sync_events()

                # Checkpoint the DB
                if not self.dont_do_pathpoints:
                    self.device.pp_database.checkpoint()
                if not self.dont_do_predicates:
                    self.device.evt_database.checkpoint()

            # Wait for variables to refresh, do we need to upload any?
            self.wait(measurement())



import enum
import typing as tp

__all__ = ['SMOKClientError', 'InvalidCredentials', 'ResponseError',
           'OperationFailedReason', 'OperationFailedError',
           'NotReadedError', 'InstanceNotReady',
           'UnavailableError']

from satella.coding.typing import Number
from satella.time import time_ms


class SMOKClientError(Exception):
    """Base class for all exceptions thrown by SMOK client"""


class InvalidCredentials(SMOKClientError):
    """
    The certificate files that you provided were invalid
    """


class InstanceNotReady(SMOKClientError):
    """
    Some operation must be done before you can call that method
    """


class UnavailableError(SMOKClientError):
    """
    The client was started in a mode that does not support your request, ie. requesting
    a BAOB while starting client with
    """


class ResponseError(SMOKClientError):
    """
    Server replied with a wrong code or didn't reply at all.
    """

    def __str__(self) -> str:
        return 'ResponseError(%s, "%s")' % (self.status_code, self.status)

    def __init__(self, status_code: tp.Optional[int] = 599, status: str = ''):
        self.status_code = status_code
        self.status = status

    def is_no_link(self) -> bool:
        """
        :return: whether this fault happened due to link failing
        """
        return self.status_code == 599

    def is_clients_fault(self) -> bool:
        """
        :return: whether this is our fault, or server's/connection's fault
        """
        return self.status_code // 100 == 4


class OperationFailedReason(enum.Enum):
    MALFORMED = 'malformed'  #: The device responded with a malformed protocol frame
    TIMEOUT = 'timeout'  #: The device did not respond within given time
    INVALID = 'invalid'  #: The device responded OK, but told us that this pathpoint is bogus
    TYPE = 'type'  #: There was an error with the typing of the value


class OperationFailedError(SMOKClientError):
    """
    Raised by the pathpoint's on_read and on_write futures when the operation fails

    :ivar reason: reason of failure
    :ivar timestamp: timestamp of failure in milliseconds
    """

    def __init__(self, reason: OperationFailedReason = OperationFailedReason.TIMEOUT,
                 timestamp: tp.Optional[Number] = None):
        self.reason = reason
        self.timestamp = timestamp or time_ms()


class NotReadedError(OperationFailedError):
    """
    The value is not available, due to it having not been yet read.

    Note that this is invalid to return in read handling futures!

    :ivar timestamp: timestamp of failure in milliseconds
    """

    def __init__(self, timestamp: tp.Optional[float] = None):
        super().__init__(None, timestamp)



import os.path
import pickle
import typing as tp

from satella.coding import Monitor
from satella.coding.typing import Number

from smok.exceptions import NotReadedError, OperationFailedError
from smok.pathpoint import PathpointValueType, ValueOrExcept
from .in_memory import InMemoryPathpointDatabase


class PicklingPathpointDatabase(InMemoryPathpointDatabase):
    """
    An example pathpoint database that persists it's data on disk.

    It persists it's store every :meth:`~smokclient.extras.BasePathpointDatabase.checkpoint` call.

    :param path: path to file containing pickled data.
    """

    def __init__(self, path: str):
        super().__init__()
        self.__path = path
        self.last_pathpoint_value = {}
        if os.path.exists(path):
            try:
                with open(path, 'rb') as f_in:
                    self.__pathpoints, self.last_pathpoint_value = pickle.load(f_in)
            except pickle.PickleError:
                pass

    def on_new_data(self, pathpoint: str, timestamp: Number,
                    value_or_exception: ValueOrExcept) -> None:
        self.last_pathpoint_value[pathpoint] = timestamp, value_or_exception
        super().on_new_data(pathpoint, timestamp, value_or_exception)

    def get_current_value(self, pathpoint: str) -> tp.Tuple[Number, PathpointValueType]:
        if pathpoint not in self.last_pathpoint_value:
            raise NotReadedError()
        val = self.last_pathpoint_value[pathpoint]
        if isinstance(val[1], OperationFailedError):
            raise val[1]
        return val

    @Monitor.synchronized
    def checkpoint(self) -> None:
        with open(self.__path, 'wb') as f_out:
            pickle.dump((self.__pathpoints, self.last_pathpoint_value), f_out)



import os
import pickle
import time
import typing as tp

from satella.coding import silence_excs

from .base import BaseMetadataDatabase


class PicklingMetadataDatabase(BaseMetadataDatabase):
    """
    Base class for metadata databases that pickles changes to disk after every change.
    """

    @silence_excs(KeyError)
    def delete_plain(self, key: str) -> None:
        del self.db_plain[key]
        self.pickle()

    def pickle(self):
        with open(self.__path, 'wb') as f_out:
            pickle.dump(self.db_plain, f_out)

    def __init__(self, path: str):
        self.db_plain = {}
        self.__path = path
        if os.path.exists(self.__path):
            try:
                with open(self.__path, 'rb') as f_in:
                    self.db_plain = pickle.load(f_in)
            except pickle.PickleError:
                pass

    def put_plain(self, key: str, value: str, timestamp: tp.Optional[float] = None) -> None:
        self.db_plain[key] = value, timestamp or time.time()
        self.pickle()

    def get_plain(self, key: str) -> str:
        return self.db_plain[key][0]

    def update_plain(self, key: str, value: str, timestamp: float) -> None:
        if key not in self.db_plain:
            self.db_plain[key] = value, timestamp
            self.pickle()
        elif self.db_plain[key][1] < timestamp:
            self.db_plain[key] = value, timestamp
            self.pickle()

    def get_all_plain(self) -> tp.Iterator[tp.Tuple[str, str, float]]:
        for key, row in self.db_plain.items():
            yield (key, *row)



import os
import pickle

from satella.coding import silence_excs
from satella.files import read_in_file, write_to_file

from .base import BaseBAOBDatabase


class PicklingBAOBDatabase(BaseBAOBDatabase):
    """
    :param path: path that has to be a directory where BAOB data will be stored.
        If that directory does not exist, it will be created
    """

    def get_baob_value(self, key: str) -> bytes:
        return read_in_file(os.path.join(self.__path, key))

    def set_baob_value(self, key: str, data: bytes, version: int):
        write_to_file(os.path.join(self.__path, key), data)
        self.versions[key] = version
        self.sync()

    def sync(self):
        with open(os.path.join(self.__path, 'metadata.pkl'), 'wb') as f_out:
            pickle.dump(self.versions, f_out, pickle.HIGHEST_PROTOCOL)

    @silence_excs(KeyError)
    def delete_baob(self, key: str) -> None:
        del self.versions[key]
        os.unlink(os.path.join(self.__path, key))
        self.sync()

    def get_baob_version(self, key: str) -> int:
        return self.versions[key]

    def __init__(self, path: str):
        self.__path = path
        if not os.path.exists(path):
            os.mkdir(path)
        try:
            with open(os.path.join(path, 'metadata.pkl', 'rb')) as f_in:
                self.versions = pickle.load(f_in)
        except (FileNotFoundError, pickle.UnpicklingError):
            self.versions = {}

    def get_all_keys(self):
        set_f = set(os.listdir(self.__path))
        with silence_excs(KeyError):
            set_f.remove('metadata.pkl')
        return iter(set_f)



import logging
import os
import socket
import ssl
import tempfile
import time
import typing as tp
from ssl import SSLContext, PROTOCOL_TLS_CLIENT, SSLError, CERT_REQUIRED

from satella.coding import silence_excs, reraise_as, Closeable, RMonitor
from satella.coding.concurrent import IDAllocator
from satella.coding.optionals import Optional
from satella.exceptions import Empty
from satella.files import read_in_file

from smok.client import get_dev_ca_cert, get_root_cert
from smok.client.certificate import get_device_info
from ..exceptions import ConnectionFailed
from ..protocol import NGTTHeaderType, STRUCT_LHH, env_to_hostname, NGTTFrame

PING_INTERVAL_TIME = 30
INTERVAL_TIME_NO_RESPONSE_KILL = 90
INTERVAL_TIME_NO_RESPONSE_READ_KILL = 120
logger = logging.getLogger(__name__)


class NGTTSocket(Closeable):
    """
    The agreement is that after last write you have 60 seconds to write something.

    We keep last_read as our-side keepalive
    """
    @property
    def wants_write(self) -> bool:
        return bool(self.w_buffer)

    def __init__(self, cert_file: str, key_file: str):
        self.monitor = RMonitor()
        self.socket = None
        self.connected = False
        environment = get_device_info(read_in_file(cert_file))[1]
        self.host = env_to_hostname(environment)
        self.cert_file = cert_file
        self.key_file = key_file
        self.buffer = bytearray()
        self.w_buffer = bytearray()
        self.ping_id = None
        self.futures = {}
        self.last_read = time.monotonic()
        self.last_write = time.monotonic()
        with tempfile.NamedTemporaryFile('wb', delete=False) as chain_file:
            chain_file.write(get_dev_ca_cert())
            chain_file.write(b'\n')
            chain_file.write(get_root_cert())
            chain_file.close()
            self.chain_file_name = chain_file.name

        self.id_assigner = IDAllocator(start_at=1, top_limit=0x8000)
        super().__init__()

    @RMonitor.synchronize_on_attribute('monitor')
    def send_frame(self, tid: int, header: NGTTHeaderType, data: bytes = b'') -> None:
        """
        Schedule a frame to be sent.

        This will be actually send in a background thread. May have a delay of
        up to 1 second, which is acceptable.

        :param tid: transaction ID
        :param header: packet type
        :param data: data to send
        """
        if self.closed:
            return
        self.w_buffer.extend(STRUCT_LHH.pack(len(data), tid, header.value))
        self.w_buffer.extend(data)

    @RMonitor.synchronize_on_attribute('monitor')
    @reraise_as((BrokenPipeError, ssl.SSLError), ConnectionFailed)
    @silence_excs(ssl.SSLWantWriteError)
    def try_send(self):
        """
        Try to send some data. This is called only when the socket is writeable.
        """
        if self.w_buffer:
            data_sent = self.socket.send(self.w_buffer)
            self.last_write = time.monotonic()
            del self.w_buffer[:data_sent]

    @RMonitor.synchronize_on_attribute('monitor')
    def try_ping(self):
        delta = time.monotonic() - self.last_write

        if time.monotonic() - self.last_read > INTERVAL_TIME_NO_RESPONSE_READ_KILL \
                and self.ping_id is not None:
            raise ConnectionFailed(False, 'timed out due to ping (read-side)')
        elif delta > INTERVAL_TIME_NO_RESPONSE_KILL \
                and self.ping_id is not None:
            raise ConnectionFailed(False, 'timed out due to ping (write-side)')
        elif delta > PING_INTERVAL_TIME and self.ping_id is None:
            try:
                self.ping_id = self.id_assigner.allocate_int()
            except Empty as e:
                logger.error('Ran out of IDs while processing ping send')
                raise ConnectionFailed(False, 'ran out of IDs on ping') from e
            logger.debug('Sending PING')
            self.send_frame(self.ping_id, NGTTHeaderType.PING)

    @RMonitor.synchronize_on_attribute('monitor')
    def got_ping(self):
        if self.ping_id is not None:
            self.id_assigner.mark_as_free(self.ping_id)
            logger.debug('Received PING')
            self.last_read = time.monotonic()
            self.ping_id = None

    def fileno(self) -> int:
        return self.socket.fileno()

    @RMonitor.synchronize_on_attribute('monitor')
    @reraise_as((BrokenPipeError, ssl.SSLError, ConnectionResetError), ConnectionFailed)
    @silence_excs(ssl.SSLWantReadError)
    def recv_frame(self) -> tp.Optional[NGTTFrame]:
        """
        Receive a frame from remote socket, or None if nothing could be assembled as of now.

        This should only be called when it is determined that the socket is readable, for example
        via :code:`select`.

        :raises ConnectionFailed: connection closed
        :return: a tuple of transaction ID, header type, data
        """
        data = self.socket.recv(512)
        if not data:
            raise ConnectionFailed(False, 'gracefully closed')
        self.last_read = time.monotonic()
        self.buffer.extend(data)

        result = NGTTFrame.from_buffer(self.buffer)
        if result is None:
            return None

        length, frame = result
        del self.buffer[:length]
        return frame

    @RMonitor.synchronize_on_attribute('monitor')
    def close(self):
        if super().close():
            self.disconnect()
            try:
                os.unlink(self.chain_file_name)
            except OSError as e:
                logger.error('Failure to remove certificate chain file %s', self.chain_file_name,
                             exc_info=e)

            for future in self.futures.values():
                future.set_exception(ConnectionFailed(False, 'closing the link'))

    @RMonitor.synchronize_on_attribute('monitor')
    def disconnect(self):
        """
        Disconnect from the remote host
        """
        with silence_excs(OSError):
            Optional(self.socket).close()
        self.socket = None
        self.connected = False

    @RMonitor.synchronize_on_attribute('monitor')
    def connect(self):
        """
        Connect to remote host

        :raises SSLError: an error occurred
        :raises RuntimeError: upon connection being closed
        """
        if self.closed:
            raise RuntimeError('This connection is closed!')
        if self.connected:
            return
        ssl_context = SSLContext(PROTOCOL_TLS_CLIENT)
        ssl_context.load_verify_locations(self.chain_file_name)
        ssl_context.load_cert_chain(self.cert_file, self.key_file)
        ssl_context.verify_mode = CERT_REQUIRED
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.settimeout(10)
        ssl_sock = ssl_context.wrap_socket(sock, server_hostname=self.host)
        try:
            ssl_sock.connect((self.host, 2408))
            ssl_sock.do_handshake()
        except ConnectionRefusedError as e:
            ssl_sock.close()
            raise ConnectionFailed(True) from e
        except (socket.error, SSLError) as e:
            ssl_sock.close()
            raise ConnectionFailed(True) from e
        self.socket = ssl_sock
        self.socket.setblocking(False)
        self.last_read = time.monotonic()
        self.last_write = time.monotonic()
        self.buffer = bytearray()
        self.w_buffer = bytearray()
        self.connected = True
        self.ping_id = None
        logger.debug('Successfully reconnected')



import os
import pickle
import typing as tp

from satella.coding import Monitor

from .base import BaseSensorDatabase
from smok.sensor.sensor import Sensor


class PicklingSensorDatabase(BaseSensorDatabase, Monitor):
    def __init__(self, path: str):
        super().__init__()
        Monitor.__init__(self)
        self.__path = path
        self.sensor_cache = {}
        if os.path.exists(path):
            try:
                with open(path, 'rb') as f_in:
                    self.__data = pickle.load(f_in)
            except pickle.UnpicklingError:
                self.__data = {}

    @Monitor.synchronized
    def get_sensor(self, fqts: str) -> Sensor:
        if fqts not in self.sensor_cache:
            sensor = Sensor(self.device, *self.__data[fqts])
            self.sensor_cache[fqts] = sensor
        return self.sensor_cache[fqts]

    @Monitor.synchronized
    def on_sensors_sync(self, sensors: tp.List[Sensor]):
        self.__data = {}
        for sensor in sensors:
            self.__data[sensor.fqts] = sensor.fqts, sensor.path, sensor.type_name
        self.__save()

    def get_all_sensors(self) -> tp.Iterator[Sensor]:
        for key in self.__data.keys():
            yield self.get_sensor(key)

    def __save(self):
        with open(self.__path, 'wb') as f_out:
            pickle.dump(self.__data, f_out, pickle.HIGHEST_PROTOCOL)



import time
import typing as tp

from satella.coding import for_argument
from satella.coding.structures import OmniHashableMixin, ReprableMixin
from satella.coding.typing import Number

from smok.pathpoint.orders import Section, ReadOrder, AdviseLevel


class ArchivingEntry(OmniHashableMixin, ReprableMixin):
    _REPR_FIELDS = ('pathpoint_name', 'interval')
    _HASH_FIELDS_TO_USE = ('pathpoint_name',)
    __slots__ = ('pathpoint_name', 'interval', 'last_updated')

    def __init__(self, pathpoint_name: str, interval: Number):
        self.pathpoint_name = pathpoint_name
        self.interval = interval
        self.last_updated = 0

    def should_update(self) -> bool:
        return time.time() - self.last_updated > self.interval

    def update(self) -> Section:
        self.last_updated = time.time()
        return Section([ReadOrder(self.pathpoint_name, AdviseLevel.ADVISE)])


@for_argument(returns=set)
def archiving_entries_from_json(y: tp.List[dict]) -> tp.Set[ArchivingEntry]:
    for entry in y:
        interval = entry['interval']
        for pp in entry['pathpoints']:
            yield ArchivingEntry(pp, interval)


def archiving_dict_from_json(y: tp.List[dict]) -> tp.Dict[int, tp.List[str]]:
    result = {}
    for entry in y:
        interval = entry['interval']
        result[entry['interval']] = entry['pathpoints']
    return result



import datetime
import io
import logging
import os
import tempfile
import threading
import time
import typing as tp
import uuid
import warnings
import weakref
from abc import ABCMeta

import pytz
from satella.coding import Closeable, for_argument, wraps
from satella.coding.concurrent import PeekableQueue, Condition
from satella.coding.optionals import Optional
from satella.coding.structures import DirtyDict
from satella.time import time_as_int

from smok.sync_workers.ngtt import NGTTSyncWorker
from smok.sync_workers.http import HTTPSyncWorker
from .api import RequestsAPI
from .certificate import get_device_info, get_root_cert, get_dev_ca_cert
from .slave import SlaveDevice
from ..baob import BAOB
from ..basics import DeviceInfo, Environment, StorageLevel
from ..exceptions import ResponseError, UnavailableError
from ..extras import BaseSensorDatabase, BaseEventDatabase, BaseMacroDatabase, \
    BasePathpointDatabase, BaseMetadataDatabase, BaseBAOBDatabase, BaseArchivesDatabase, \
    BaseSensorWriteDatabase, BasePredicateDatabase
from ..extras.arch_database.in_memory import InMemoryArchivesDatabase
from ..extras.baob_database.memory import InMemoryBAOBDatabase
from ..extras.event_database import InMemoryEventDatabase
from ..extras.macros_database.in_memory import InMemoryMacroDatabase
from ..extras.metadata_database import InMemoryMetadataDatabase
from ..extras.pp_database.in_memory import InMemoryPathpointDatabase
from ..extras.pred_database.in_memory import InMemoryPredicateDatabase
from ..extras.sensor_write_database.in_memory import InMemorySensorWriteDatabase
from ..extras.sensors_database.in_memory import InMemorySensorDatabase
from ..metadata import PlainMetadata
from ..pathpoint import Pathpoint, ReparsePathpoint
from ..pathpoint.orders import Section, MessageOrder
from ..predicate import BaseStatistic, Event, Color
from ..predicate.registration import CollectionOfStatistics, StatisticRegistration
from ..predicate.undefined import UndefinedStatistic
from ..sensor import Sensor, fqtsify, SensorWriteEvent
from ..threads import OrderExecutorThread, CommunicatorThread, ArchivingAndMacroThread, \
    LogPublisherThread
from ..threads.communicator import PREDICATE_SYNC_INTERVAL

logger = logging.getLogger(__name__)


def must_be_open(fun):
    @wraps(fun)
    def inner(self, *args, **kwargs):
        if self.closed:
            raise RuntimeError('SMOK device already closed!')
        return fun(self, *args, **kwargs)
    return inner


class SMOKDevice(Closeable, metaclass=ABCMeta):
    """
    A base class for a SMOK device.

    You should subclass it, to provide your own device.

    Note that instantiating this object spawns two non-daemon thread. This object must be
    close()d before termination (or garbage collected).

    :param cert: either a path to or a file-like object containing the device certificate
    :param priv_key: either a path to or a file-like object containing the device private key
    :param allow_sync: whether to allow outbound TCP communication to automatically synchronize
        background. Note that if you explicitly call a method that calls TCP this does not apply.
        Can be lated changed.
    :param evt_database: custom event database. Providing a string defaults to path where predicate
        data will be persisted. Since events are persistent, a store needs to be given.
    :param pp_database: custom pathpoint value database. Default value of None defaults to an
        in-memory implementation
    :param macro_database: custom macro database. Default value of None will result in an
        in-memory implementation
    :param meta_database: custom meta database. Default value of None will result in an
        in-memory implementation
    :param sensor_database: custom sensor database. Default value of None will result in an
        in-memory implementation
    :param baob_database: custom BAOB database. Default value of None will result in an
        in-memory implementation
    :param pred_database: custom Predicate database. Default value of None will result in an
        in-memory implementation
    :param arch_database: custom archives database. Default value of None will result in an
        in-memory implementation
    :param sensor_write_database: custom sensor write database. Default value of None will result in
        an in-memory implementation
    :param dont_obtain_orders: if set to True, this SMOKDevice won't poll for orders.
        This also implies dont_do_baobs. It is a ValueError to set this while setting
        dont_bo_baobs to False.
    :param dont_do_pathpoints: if set to True, this SMOKDevice won't support pathpoints
        or sensors.
    :param dont_sync_sensor_writes: if set to True, sensor writes won't be synced
    :param dont_do_baobs: if set to True, this SMOKDevice won't care about BAOBs.
    :param dont_do_macros: if set to True, this SMOKDevice won't take care of the macros
    :param dont_do_predicates: if set to True, this SMOKDevice won't do predicates
    :param dont_do_archives: if set to True, this SMOKDevice won't do archiving
    :param startup_delay: amount of seconds to wait after creation for CommunicatorThread to
        start talking and OrderExecutorThread to start grabbing orders. Deprecated.
    :param cache_metadata_for: amount of seconds to cache downloaded metadata entry.
        Ie no attempt to download them from the server again will be made in that many
        seconds since the download.
    :param delayed_boot: if set to True, you will need to call
        :meth:`~smok.client.SMOKDevice.continue_boot` in order to start fetching orders and the
        like. False by default.
    :param use_ngtt: if set to True, logs, orders and pathpoint data will be transmitted over
        a persistent TLS connection instead of HTTP API.


    About 10 seconds from creation if CommunicatorThread was created, sensors will be synced and
    the device will start talking. To reduce this delay, set parameter startup_delay

    If both dont_do_macros and dont_do_archives are True, the archiving & macro thread
    won't be started.
    If dont_obtain_orders is True, then order executor and order getter threads won't be started.

    Every method is supposed to be called from user threads (SMOKDevice's service threads included),
    unless noted otherwise.

    :ivar device_id: device ID of this device (str)
    :ivar allow_sync: whether to allow background synchronization (bool).
        This can be changed by setting it directly to correct value. Setting this to False
        will help you save some bandwidth. Note that log upload will be put on hold, until it
        overflows the buffer. In that case, earlier submitted logs will be thrown on the floor
        to make room for new entries.
    :ivar pathpoints: a dictionary, keying pathpoint names to their instances
    :ivar url: base URL for the API calls, without the trailing slash
    :ivar metadata: plain metadata for this device
        (class :class:`smok.metadata.PlainMetadata`)
    :ivar baobs_loaded: whether all BAOBS have been synchronized (bool)
    :ivar cert_data: (bytes) device certificate as given by the user

    :raise ValueError: invalid combination of arguments was given
    """

    def provide_unknown_pathpoint(self, name: str,
                                  storage_level: StorageLevel = StorageLevel.TREND) -> Pathpoint:
        """
        Override this method to generate pathpoints that are referred to by incoming commands,
        but not defined yet.

        The default implementation always raises `KeyError`.

        .. note:: this can safely raise `KeyError` upon encountering a predicate that is manually
                  defined and registered via :class:`~smokclient.pathpoint.Pathpoint`

        :return: a Pathpoint instance corresponding to what was ordered
        :raises KeyError: pathpoint could not be generated
        """
        return Pathpoint(self, name, storage_level)

    def sync_sections(self, termination_checker: tp.Callable[[], bool]):
        """
        Called by order executor thread before a CANNOT_JOIN section is about to be executed.

        Override this method if you are providing a custom
        :meth:`~smok.client.SMOKDevice.execute_section`. Default does nothing.

        This needs to block until all orders issued up to this point are finished.

        Called by the order executor thread, that's why it can safely block.

        Sections should be synchronized in bounded time.

        :param termination_checker: a callable that can be called to determine whether given
            SMOKDevice is undergoing a shutdown. If this callable starts to return True,
            the function should return at once.
        """

    def execute_section(self, section: Section) -> None:
        """
        Override to implement custom section execution.

        If this is not overridden, standard executor logic will be applied.

        This will not provide you with cancelled sections. You also do not need to mark
        sections as complete, as this is done by the executor thread.

        Called by the order executor thread. If this is defined, then
        :meth:`~smok.client.SMOKDevice.sync_sections` is necessary as well, so this should just
        execute the orders themselves.

        :param section: section to execute
        """

    def __init__(self, cert: tp.Union[str, io.StringIO],
                 priv_key: tp.Union[str, io.StringIO],
                 evt_database: tp.Union[str, BaseEventDatabase],
                 pp_database: tp.Optional[BasePathpointDatabase] = None,
                 macro_database: tp.Optional[BaseMacroDatabase] = None,
                 meta_database: tp.Optional[BaseMetadataDatabase] = None,
                 sensor_database: tp.Optional[BaseSensorDatabase] = None,
                 baob_database: tp.Optional[BaseBAOBDatabase] = None,
                 pred_database: tp.Optional[BasePredicateDatabase] = None,
                 arch_database: tp.Optional[BaseArchivesDatabase] = None,
                 sensor_write_database: tp.Optional[BaseSensorWriteDatabase] = None,
                 allow_sync: bool = True,
                 dont_obtain_orders: bool = False,
                 dont_sync_sensor_writes: bool = False,
                 dont_do_macros: bool = False,
                 dont_do_predicates: bool = False,
                 dont_do_pathpoints: bool = False,
                 dont_do_baobs: bool = False,
                 dont_do_archives: bool = False,
                 cache_metadata_for: float = 60,
                 startup_delay: tp.Optional[float] = None,
                 delayed_boot: bool = False,
                 use_ngtt: bool = False):
        if startup_delay is not None:
            warnings.warn('This is deprecated. Use delayed_boot', DeprecationWarning)
        else:
            startup_delay = 0
        super().__init__()
        Optional(baob_database).check_consistency()
        self.cache_metadata_for = cache_metadata_for
        self.dont_do_predicates = dont_do_predicates
        self.dont_do_pathpoints = dont_do_pathpoints
        self.dont_sync_sensor_writes = dont_sync_sensor_writes
        self.pp_database = pp_database or InMemoryPathpointDatabase()
        self.baobs_loaded = False
        self.allow_sync = allow_sync
        if isinstance(evt_database, str):
            self.evt_database = InMemoryEventDatabase(evt_database)
        else:
            self.evt_database = evt_database
        self.macros_database = macro_database or InMemoryMacroDatabase()
        self.meta_database = meta_database or InMemoryMetadataDatabase()
        self.pred_database = pred_database or InMemoryPredicateDatabase()
        self.delayed_boot = delayed_boot
        self.sensor_database = sensor_database or InMemorySensorDatabase()
        self.sensor_database.on_register(self)
        self.use_ngtt = use_ngtt
        self.arch_database = arch_database or InMemoryArchivesDatabase()
        self.baob_database = baob_database or InMemoryBAOBDatabase()
        self.sensor_write_database = sensor_write_database or InMemorySensorWriteDatabase()
        self.metadata = PlainMetadata(self)
        self.ready_lock = threading.Lock()
        self.ready_lock.acquire()

        self.predicates = {}  # type: tp.Dict[str, BaseStatistic]

        # Load cached predicates
        for predicate in self.pred_database.get_all_predicates():
            udf = UndefinedStatistic(device=self, **predicate)
            self.predicates[udf.predicate_id] = udf

        self._timezone = None
        self.statistic_registration = CollectionOfStatistics()
        self._statistics_updated = False
        self.pathpoints = DirtyDict()  # type: tp.Dict[str, Pathpoint]
        self.temp_file_for_cert = None
        self.cert_file_name = None
        if not isinstance(cert, str):
            with tempfile.NamedTemporaryFile('w', delete=False) as cert_file:
                cert_file.write(cert.read())
            cert = self.temp_file_for_cert = cert_file.name
        else:
            self.cert_file_name = self.temp_file_for_cert = cert

        self.priv_key_file_name = None
        self.temp_file_for_key = None
        if not isinstance(priv_key, str):
            with tempfile.NamedTemporaryFile('w', delete=False) as key_file:
                key_file.write(priv_key.read())
            priv_key = self.temp_file_for_key = key_file.name
        else:
            self.priv_key_file_name = self.temp_file_for_key = priv_key

        self.cert = cert, priv_key

        with open(cert, 'rb') as fin:
            cert_data = fin.read()

        self.cert_data = cert_data
        dev_id, env = get_device_info(cert_data)
        self.device_id = dev_id  # type: str
        self.environment = env  # type: Environment
        if self.environment == Environment.PRODUCTION:
            self.url = 'https://api.smok.co'
        elif self.environment == Environment.STAGING:
            self.url = 'https://api.test.smok-serwis.pl'
        elif self.environment == Environment.LOCAL_DEVELOPMENT:
            self.url = 'http://http-api'

        self.api = RequestsAPI(self)
        self.log_publisher = LogPublisherThread(self)

        self._order_queue = PeekableQueue()
        if not (dont_do_archives and dont_do_macros):
            self.arch_and_macros = ArchivingAndMacroThread(self, self._order_queue,
                                                           dont_do_macros, dont_do_archives)
            if not delayed_boot:
                self.arch_and_macros.start()
        else:
            self.arch_and_macros = None
        self.dont_do_baobs = dont_do_baobs
        self.dont_do_orders = dont_obtain_orders
        self.executor = None
        self.getter = None
        if not dont_obtain_orders or not dont_do_predicates or not dont_do_pathpoints or \
                not dont_do_baobs:
            self.executor = OrderExecutorThread(self, self._order_queue, self.pp_database,
                                                startup_delay)
            self.getter = CommunicatorThread(self, self._order_queue, self.pp_database,
                                             dont_obtain_orders,
                                             dont_do_baobs,
                                             dont_do_pathpoints,
                                             dont_do_predicates,
                                             dont_sync_sensor_writes, startup_delay)
        if not delayed_boot:
            if self.use_ngtt:
                self.sync_worker = NGTTSyncWorker(self)
            else:
                self.sync_worker = HTTPSyncWorker(self)
            if not dont_obtain_orders or not dont_do_predicates or not dont_do_pathpoints or \
                    not dont_do_baobs:
                self.executor.start()
                self.getter.start()
            self.log_publisher.start()
            self.boot_completed = True
        else:
            self.sync_worker = None
            self.boot_completed = False

    def continue_boot(self):
        """
        Call this to continue the booting if delayed_start was given in the constructor

        This will start the communicator and order executor thread.

        :raise RuntimeError: delayed boot was not given
        """
        if not self.delayed_boot:
            raise RuntimeError('Delayed boot was not given')
        if self.use_ngtt:
            self.sync_worker = NGTTSyncWorker(self)
        else:
            self.sync_worker = HTTPSyncWorker(self)
        Optional(self.executor).start()
        Optional(self.getter).start()
        Optional(self.arch_and_macros).start()
        self.log_publisher.start()
        self.boot_completed = True

    @must_be_open
    def log_sensor_write(self, sw: SensorWriteEvent):
        """
        Log that a sensor has been written and enqueue it for cloud upload

        :param sw: sensor write event to upload
        :raises RuntimeError: device already closed
        """
        self.sensor_write_database.add_sw(sw)

    @must_be_open
    def reset_predicates(self):
        """
        Clear all loaded predicates and force a renew of loading.

        Discards all currently loaded Predicate instances. Currently registered statistics will
        remain registered.

        .. warning:: Currently requires Internet access to restore predicates
        :raises RuntimeError: device already closed
        """
        self.predicates = {}
        self.getter.last_predicates_synced = time.monotonic() - PREDICATE_SYNC_INTERVAL

    @must_be_open
    def get_baob(self, key: str) -> BAOB:
        """
        Retrieve given BAOB

        :raises UnavailableError: client was launched in a mode with BAOBs disabled
        :raises RuntimeError: device already closed
        """
        if self.dont_do_baobs:
            raise UnavailableError('Support for BAOBs was disabled')
        return BAOB(self, key)

    @must_be_open
    def get_all_baobs(self) -> tp.Iterator[BAOB]:
        """
        Stream all BAOBs

        :raises UnavailableError: client was launched in a mode with BAOBs disabled
        :raises RuntimeError: device already closed
        """
        if self.dont_do_baobs:
            raise UnavailableError('Support for BAOBs was disabled')

        for key in self.baob_database.get_all_keys():
            yield BAOB(self, key)

    def wait_until_synced(self) -> None:
        """
        Block until everything's synchronized with the server
        """
        self.ready_lock.acquire()
        self.ready_lock.release()

    def _execute_message_order(self, order: MessageOrder) -> None:
        """
        Tell the server to execute provided :class:`~smok.pathpoint.orders.MessageOrder`

        :param order: order to execute
        :meta public:
        """
        for i in range(3):
            try:
                self.api.post('/v1/device/orders/message/' + order.uuid)
                self.on_successful_sync()
            except ResponseError as e:
                if e.is_no_link():
                    self.on_failed_sync()

    @property
    def timezone(self) -> pytz.timezone:
        """
        :return: the timezone this device is in
        """
        if self._timezone is None:
            self.get_device_info()
        return pytz.timezone(self._timezone)

    @must_be_open
    def close_event(self, event: Event, timestamp: tp.Optional[int] = None) -> None:
        """
            Close the provided event

        :param event: event to close
        :param timestamp: timestamp of close. Defaults to now
        :raises RuntimeError: device already closed
        """
        assert not event.is_closed()
        if event.ended_on is None:
            event.ended_on = timestamp or time_as_int()
        self.evt_database.close_event(event)

    @must_be_open
    def get_open_event(self, event_id: str) -> Event:
        """
        Return a particular opened event

        :param event_id: opened event UUID
        :return: a particular event
        :raises KeyError: event not found, or already closed
        :rtype: Event
        :raise UnavailableError: SMOKDevice was launched in a no-predicate mode
        :raises RuntimeError: device already closed
        """
        if self.dont_do_predicates:
            raise UnavailableError('SMOKDevice was launched without predicates')
        for event in self.evt_database.get_open_events():
            if event.uuid_matches(event_id):
                return event
        raise KeyError()

    @must_be_open
    def get_all_sensors(self) -> tp.Iterator[Sensor]:
        """
        Stream all sensors

        .. note:: This will block until sensors are synced from the server

        :raise UnavailableError: SMOKDevice was launched in a no-pathpoint mode
        :raises RuntimeError: device already closed
        """
        if self.dont_do_pathpoints:
            raise UnavailableError('SMOKDevice was launched without pathpoints')
        with self.ready_lock:
            yield from self.sensor_database.get_all_sensors()

    def on_baob_updated(self, baob_name: str) -> None:
        """
        Called by CommunicatorThread after given BAOB was updated.

        After this is called, the new BAOB can be successfully loaded and it's new
        contents will be retrieved.

        This is not invoked during the first synchronization, nor after you change the BAOB
        on the client side.

        :param baob_name: name of the BAOB that was just downloaded from the server
        """

    @must_be_open
    def get_sensor(self, tag_set: tp.Union[tp.Set[str], str]) -> Sensor:
        """
        Return a sensor

        .. note:: This will block until sensors are synced from the server

        :param tag_set: either set of strs or these strs joined with a ' '
        :return: sensor
        :raises KeyError: sensor does not exist
        :raise UnavailableError: SMOKDevice was launched in a no-pathpoint mode
        :raises RuntimeError: device already closed
        """
        if self.dont_do_pathpoints:
            raise UnavailableError('SMOKDevice was launched without pathpoints')
        with self.ready_lock:
            if isinstance(tag_set, set):
                tag_set = list(tag_set)
                tag_set.sort()
                tag_set = ' '.join(tag_set)
            else:
                tag_set = fqtsify(tag_set)
            return self.sensor_database.get_sensor(tag_set)

    @must_be_open
    def get_all_events(self) -> tp.Iterator[Event]:
        """
        Return all events kept in device's database

        :raise UnavailableError: SMOKDevice was launched in a no-predicate mode
        :raises RuntimeError: device already closed
        """
        if self.dont_do_predicates:
            raise UnavailableError('SMOKDevice was launched without predicates')
        return self.evt_database.get_all_events()

    @must_be_open
    def get_all_open_events(self) -> tp.Iterator[Event]:
        """
        Get all open events

        :raise UnavailableError: SMOKDevice was launched in a no-predicate mode
        :raises RuntimeError: device already closed
        """
        if self.dont_do_predicates:
            raise UnavailableError('SMOKDevice was launched without predicates')
        return self.evt_database.get_open_events()

    @must_be_open
    def open_event(self, started_on: int, ended_on: tp.Optional[int],
                   color: Color, is_point: bool, token: str, group: str, message: str,
                   metadata: tp.Optional[tp.Dict[str, str]] = None) -> Event:
        """
        Create a new event

        :param started_on: timestamp in seconds, when was the event started?
        :param ended_on: timestamp in seconds when has the event ended, None in case of
            open events.
        :param color: :term:`Color` of an event
        :param is_point: whether this is a :term:`point event`
        :param token: a string
        :param group: notification group
        :param message: human-readable message
        :param metadata: extra metadata. This must be dict'able
        :return: the Event object
        :raise UnavailableError: SMOKDevice was launched in a no-predicate mode
        :raises RuntimeError: device already closed
        """
        if self.dont_do_predicates:
            raise UnavailableError('SMOKDevice was launched without predicates')
        metadata = dict(metadata or {})
        evt_uuid = uuid.uuid4().hex
        event = Event(evt_uuid, started_on, ended_on, color, is_point, token, group, message,
                      None, metadata)
        logger.warning('Got %s of type %s', event, type(event))
        self.evt_database.add_event(event)
        return event

    def execute(self, *secs: Section) -> None:
        """
        Schedule sections to be executed.

        To be invoked by any thread. Use this to inject sections into device's execution loop.

        :param secs: sections to be executed, in that order, unless they're joinable
        """
        for sec in secs:
            self._order_queue.put(sec)

    @must_be_open
    def get_pathpoint(self, path: str,
                      storage_level: StorageLevel = StorageLevel.TREND) -> Pathpoint:
        """
        Obtain a pathpoint. Creates one and registers it if not available.

        :param path: path of the pathpoint
        :param storage_level: target storage level
        :return: a pathpoint having provided name
        :raises KeyError: pathpoint not available
        :raise UnavailableError: SMOKDevice was launched in a no-pathpoint mode
        :raises RuntimeError: device already closed
        """
        if self.dont_do_pathpoints:
            raise UnavailableError('SMOKDevice was launched without pathpoints')
        if path[0] == 'r':
            return ReparsePathpoint(self, path, storage_level)
        if path in self.pathpoints:
            return self.pathpoints[path]
        pp = self.provide_unknown_pathpoint(path, storage_level)  # raises KeyError
        self.register_pathpoint(pp)
        return pp

    @must_be_open
    def register_statistic(self, stat: tp.Type[BaseStatistic],
                           predicate: tp.Callable[[str, dict], bool]) -> StatisticRegistration:
        """
        Register a new statistic.

        Statistics can be registered at any point. If there are pending predicates,
        instances will be created for them shortly by the communicator thread.

        :param stat: a class (not an instance) to register
        :param predicate: a callable taking two arguments: statistic name and it's configuration.
            The callable should return whether to apply stat to this predicate
        :return: a Registration object. Can be later cancelled.
        :raise UnavailableError: SMOKDevice was launched in a no-predicate mode
        :raises RuntimeError: device already closed
        """
        if self.dont_do_predicates:
            raise UnavailableError('SMOKDevice was launched without predicates')
        assert issubclass(stat, BaseStatistic), 'Not a subclass of BaseStatistic!'
        reg = StatisticRegistration(predicate, stat)
        self.statistic_registration.add(reg)
        return reg

    @must_be_open
    def register_pathpoint(self, pp: Pathpoint) -> None:
        """
        Register a pathpoint for usage with this SMOKDevice.

        Normally, you shouldn't need to use it, as the :class:`~smokclient.pathpoint.Pathpoint`
        constructor does that for you. However, if you provide it's `SMOKDevice` parameter as None,
        this call is still required

        :param pp: pathpoint to register
        :raise UnavailableError: SMOKDevice was launched in a no-pathpoint mode
        :raises RuntimeError: device already closed
        """
        if self.dont_do_pathpoints:
            raise UnavailableError('SMOKDevice was launched without pathpoints')
        if pp.name[0] == 'r':
            return
        if pp.name not in self.pathpoints:
            pp.device = weakref.proxy(self)
            self.pathpoints[pp.name] = pp

    def execute_sysctl(self, op_type: str, op_args: str) -> bool:
        """
        Called by executor thread upon receiving a request to execute a particular SysctlOrder

        Handles commonly defined BAOBs. Extend to implement custom sysctls.
        Sysctl orders are user-defined.

        :param op_type: type of operation to execute
        :param op_args: argument of the operation to execute.
        :return: whether this command was recognized and acted upon
        """
        if self.dont_do_baobs:
            return False
        if op_type in ('baob-updated', 'baob-created'):
            if self.getter is not None:
                self.getter.last_baob_synced = 0
                self.getter.data_to_update.notify()
            return True
        elif op_type == 'baob-deleted':
            self.baob_database.delete_baob(op_args)
            return True
        return False

    def close(self) -> None:
        """
        Close the connection, clean up the resources.

        This may block for up to 10 seconds.

        No-op if called more than once.
        """
        if super().close():
            Optional(self.executor).terminate()
            Optional(self.getter).terminate()
            self.log_publisher.terminate()
            Optional(self.arch_and_macros).terminate()
            if self.priv_key_file_name is None:
                os.unlink(self.temp_file_for_key)
            if self.cert_file_name is None:
                os.unlink(self.temp_file_for_cert)
            Optional(self.executor).join()
            Optional(self.getter).join()
            if self.boot_completed:
                self.log_publisher.join()
            Optional(self.arch_and_macros).join()

    @must_be_open
    @for_argument(returns=list)
    def get_slaves(self) -> tp.List[SlaveDevice]:
        """
        Return information about slave devices

        :return: a list of slave devices
        :raises ResponseError: server responded (or not) with an invalid message
        :raises RuntimeError: device already closed
        """
        slaves = self.get_device_info().slaves
        for slave in slaves:
            yield SlaveDevice(self, slave)

    @must_be_open
    def get_device_info(self) -> DeviceInfo:
        """
        Obtain information about the device.

        Note that this will result in :meth:`~smok.client.SMOKDevice.on_failed_sync`
        or :meth:`~smok.client.SMOKDevice.on_successful_sync` being called.

        :return: current device information
        :raises ResponseError: server responded (or not) with an invalid message
        :raises RuntimeError: :attr:`~smok.client.SMOKDevice.allow_sync` was set to False or
            device already closed
        """
        if not self.allow_sync:
            raise RuntimeError('allow_sync is False, cannot fetch the information')
        try:
            resp = DeviceInfo.from_json(self.api.get('/v1/device'))
        except ResponseError as e:
            if e.is_no_link():
                self.on_failed_sync()
            raise
        self.on_successful_sync()
        self._timezone = resp.timezone
        return resp

    def on_successful_sync(self, *args, **kwargs) -> None:
        """
        Called by CommunicatorThread each time a part of the system synchronizes correctly with the
        server. Can be used to implement link liveness detector.

        args and kwargs left for future extendability.

        Please note that if you set allow_sync to False this won't be called, as the client won't
        try to talk to the server. The only chance is getting synces manually, as via
        :meth:`~smok.client.SMOKDevice.get_device_info`
        """

    def on_failed_sync(self, *args, **kwargs) -> None:
        """
        Called by CommunicatorThread each time a part of the system fails to synchronize by
        receiving no response at all from the server. Can be used to implement link liveness
        detector.

        args and kwargs left for future extendability

        Please note that if you set allow_sync to False this won't be called, as the client won't
        try to talk to the server. The only chance is getting synces manually, as via
        :meth:`~smok.client.SMOKDevice.get_device_info`
        """

    @must_be_open
    def get_local_time(self) -> datetime.datetime:
        """
        Return current local time on target culture context

        :return: a datetime object having the local time for this device
        :raises RuntimeError: device already closed
        """
        # What is the time on target device?
        tz = self.timezone

        utc_time = pytz.UTC.localize(datetime.datetime.utcfromtimestamp(time.time()))
        local_time = utc_time.astimezone(tz)

        return local_time

    @property
    def cert_chain(self) -> bytes:
        """
        Return your own certificate chain, finishing at SMOK CA certificate

        :return: certificate chain in PEM format
        """
        my_dat = self.cert_data
        dev_ca_dat = get_dev_ca_cert()
        root_ca_dat = get_root_cert()
        return b''.join((my_dat, dev_ca_dat, root_ca_dat))



import logging
import typing as tp

from satella.coding.transforms import merge_series
from satella.coding.typing import Number
from satella.exceptions import ImpossibleError

from .orders import AdviseLevel, Section
from .typing import PathpointValueType, ValueOrExcept
from .pathpoint import Pathpoint
from ..basics import StorageLevel
from ..exceptions import OperationFailedError
from ..sensor.reparse import parse
from ..sensor.reparse_funs import ecre_eval

logger = logging.getLogger(__name__)


class ReparsePathpoint(Pathpoint):
    """
    A pathpoint whose value is derived from other pathpoints.

    It's first letter is 'r' then there goes the type letter and a Python expression
    with other pathpoints substituted by braces, so eg. rf{W1r4002}+2 would be
    pathpoint "W1r4002" plus two.

    Writing a reparse pathpoint always results in a zero operation, while reading
    it reads all constituent pathpoints.

    Good news is that you don't need to override it, nor provide your own implementations.
    """
    __slots__ = 'expr', 'slave_pathpoints'

    def __init__(self, device: tp.Optional['SMOKDevice'], name: str,
                 storage_level: StorageLevel = StorageLevel.TREND):
        super().__init__(device, name, storage_level)
        self.expr, pps = parse(name[2:])
        self.slave_pathpoints = []
        for pp in pps:
            self.slave_pathpoints.append(device.get_pathpoint(pp, storage_level))

    def on_new_storage_level(self, new_storage_level: StorageLevel) -> None:
        """
        Called when this pathpoint's storage level is updated.

        :param new_storage_level: new storage level for this pathpoint
        """
        for slave in self.slave_pathpoints:
            new_storage_level = slave.storage_level or new_storage_level
            if new_storage_level != slave.storage_level:
                slave.on_new_storage_level(new_storage_level)
        self.storage_level = new_storage_level

    def get_archive(self,
                    starting_at: int,
                    stopping_at: tp.Optional[int] = None) -> tp.Iterator[
        tp.Tuple[int, ValueOrExcept]]:
        """
        Get archive readings.

        This will be readed entirely from the device's data, the server will not be queried

        :param starting_at: timestamp of start, in milliseconds
        :param stopping_at: timestamp of end, in milliseconds, or None for the end of the park
        :return: an iterator of tuple (timestamp in milliseconds,
                                       sensor value or exception instance)
        """
        slave_archives = []
        for slave in self.slave_pathpoints:
            slave_archives.append(slave.get_archive(starting_at, stopping_at))
        merged_series = merge_series(*slave_archives)
        for ts, *values in merged_series:
            for v in values:
                if isinstance(v, OperationFailedError):
                    yield ts, v
                    break
            else:
                yield ts, ecre_eval(self.expr, args=values)

    def set_new_value(self, timestamp: Number, value: ValueOrExcept) -> None:
        """
        A no-op
        """

    def get(self) -> tp.Tuple[Number, PathpointValueType]:
        """
        Calculate the reparse pathpoint's value

        :return: a tuple of (timestamp in milliseconds, value)
        :raises OperationFailedError: when any constituent's pathpoint's read has failed
        """
        timestamps, values = [], []
        for slave in self.slave_pathpoints:
            ts, v = slave.get()
            timestamps.append(ts)
            values.append(v)
        return max(timestamps), ecre_eval(self.expr, args=values)

    def write(self, value, advise_level=AdviseLevel.ADVISE,
              stale_after=None) -> Section:
        """
        Writing a reparse pathpoint is always a zero operation
        """
        return Section()

    def read(self, advise_level=AdviseLevel.ADVISE) -> Section:
        """
        Reading a reparse pathpoint results in all of it's constituents being read.

        :return: a Section, whose execution will result in this pathpoint being re-read
        """
        sec = Section()
        for slave in self.slave_pathpoints:
            sec += slave.read(advise_level)
        return sec

    def on_read(self, advise) -> tp.NoReturn:
        """
        It is an ImpossibleError to call this
        """
        raise ImpossibleError('should never be called!')

    def on_write(self, value, advise) -> tp.NoReturn:
        """
        It is an ImpossibleError to call this
        """
        raise ImpossibleError('should never be called!')



import typing as tp

from .base import BaseStatistic


class StatisticRegistration:
    """
    A proof that given statistic has been registered
    """
    __slots__ = ('matcher', 'stat_class', 'cancelled')

    def __init__(self, matcher: tp.Callable[[str, tp.Dict], bool],
                 stat_class: tp.Type[BaseStatistic]):
        self.matcher = matcher
        self.stat_class = stat_class
        self.cancelled = False

    def cancel(self):
        """
        Cancel this registration.

        Note that this won't update existing predicates.
        """
        self.cancelled = True

    def matches(self, stat_name: str, configuration: tp.Dict):
        if self.cancelled:
            return False
        return self.matcher(stat_name, configuration)


class CollectionOfStatistics:
    __slots__ = ('registrations',)

    def __init__(self):
        self.registrations = []

    def clear(self):
        self.registrations = []

    def clear_cancelled(self):
        self.registrations = [reg for reg in self.registrations if not reg.cancelled]

    def add(self, reg: StatisticRegistration):
        self.registrations.append(reg)

    def try_match(self, *args) -> tp.Optional[tp.Type[BaseStatistic]]:
        if len(args) == 1:
            args = args[0].statistic, args[0].configuration
        self.clear_cancelled()
        reg: StatisticRegistration
        for reg in self.registrations:
            if reg.matches(*args):
                return reg.stat_class
        return None
